import os
from tempfile import TemporaryDirectory
from io import StringIO
from rkd.inputoutput import BufferedSystemIO
from rkd.inputoutput import IO
from rkd_cooperative import CooperativeSyncTask
from rkd_cooperative.test import BaseTestCase


class CooperativeSyncTaskTest(BaseTestCase):
    def test_functional_syncing_repository_is_cloning_then_doing_pull_next_time(self):
        """Functional test checking if there is a git clone and git pull performed"""

        with TemporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)
            io = BufferedSystemIO()
            r_io = IO()
            str_io = StringIO()

            task: CooperativeSyncTask = CooperativeSyncTask()
            self.satisfy_task_dependencies(task=task, io=io)

            args = {
                '--repositories': 'https://github.com/riotkit-org/riotkit-harbor-snippet-cooperative' +
                                       '@@791f617e664f3d0d50d485b88abe38b87cbb2525'
            }

            with r_io.capture_descriptors(enable_standard_out=True, stream=str_io):
                # first time, it should be `git clone`
                task.execute(self.mock_execution_context(task, args=args,
                                                         defined_args={'--repositories': {'default': ''}}))

                # second time it should be a `git checkout && git pull`
                task.execute(self.mock_execution_context(task, args=args,
                                                         defined_args={'--repositories': {'default': ''}}))

            content = str_io.getvalue()
            self.assertIn("Cloning into '.rkd/cooperative/riotkit-org/riotkit-harbor-snippet-cooperative'", content,
                          msg='Expected a git clone')
            self.assertIn("HEAD is now at 791f617 Modify structure", content,
                          msg='Expected checkout on branch')
            self.assertIn('791f617e664f3d0d50d485b88abe38b87cbb2525 -> FETCH_HEAD', content,
                          msg='Expected a pull on a commit')
            self.assertIn('Already up to date.', content,
                          msg='Expected a checkout')

    def test_extract_repository_name_from_git_url(self):
        tested_method = CooperativeSyncTask.extract_repository_name_from_git_url

        self.assertEqual('riotkit-org/riotkit-do', tested_method('https://github.com/riotkit-org/riotkit-do'))
        self.assertEqual('riotkit-org/riotkit-do', tested_method('https://github.com:5000/riotkit-org/riotkit-do'))
        self.assertEqual('riotkit-org/riotkit-do', tested_method('git@github.com:riotkit-org/riotkit-do.git'))
        self.assertEqual('riotkit-org/riotkit-do', tested_method('ssh://git@github.com:riotkit-org/riotkit-do.git'))
        self.assertEqual('riotkit-org/riotkit-do', tested_method('ssh://git@github.com:5000/riotkit-org/riotkit-do.git'))

    def test_functional_next_repository_will_be_cloned_even_if_first_fails(self):
        """Verify that repository "B" will be cloned even if "A" failed"""

        with TemporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)

            io = BufferedSystemIO()
            task: CooperativeSyncTask = CooperativeSyncTask()
            self.satisfy_task_dependencies(task=task, io=io)

            args = {
                '--repositories': 'https://github.com/riotkit-org/riotkit-harbor-snippet-cooperative' +
                                       '@@invalid-ref-name,' +
                                       'https://github.com/riotkit-org/empty'
            }

            result = task.execute(self.mock_execution_context(task, args=args))

            self.assertTrue(os.path.isdir('.rkd/cooperative/riotkit-org/empty'),
                            msg='Expected empty directory to be cloned for testing')
            self.assertIn('Failed to synchronize repository ' +
                          '"https://github.com/riotkit-org/riotkit-harbor-snippet-cooperative"', io.get_value())
            self.assertFalse(result)

    # todo: Cover case - cloning a local git repository from directory to directory



import os
from copy import deepcopy
from abc import ABC
from rkd.api.testing import BasicTestingCase


class BaseTestCase(BasicTestingCase, ABC):
    _original_cwd: str
    _original_environ: dict

    def setUp(self) -> None:
        self._original_cwd = os.getcwd()
        self._original_environ = deepcopy(os.environ)

    def tearDown(self) -> None:
        os.chdir(self._original_cwd)
        os.environ = self._original_environ



import os
import subprocess
from glob import glob
from argparse import ArgumentParser
from typing import Dict
from typing import Tuple
from urllib.parse import urlparse
from rkd.api.contract import ExecutionContext
from rkd.api.inputoutput import Wizard
from rkd.api.contract import TaskInterface
from rkd.api.contract import ArgumentEnv
from rkd.exception import MissingInputException
from .formatting import core_snippet_tasks_formatting

HARBOR_PATH = os.path.dirname(os.path.realpath(__file__)) + '/..'
REPOSITORIES_DIRECTORY = '.rkd/cooperative'


class BaseCooperativeTask(TaskInterface):
    def format_task_name(self, name) -> str:
        return core_snippet_tasks_formatting(name)

    def get_declared_envs(self) -> Dict[str, ArgumentEnv]:
        envs = super(BaseCooperativeTask, self).get_declared_envs()
        envs['COOP_REPOSITORIES'] = ArgumentEnv(name='COOP_REPOSITORIES', switch='--repositories', default='')

        return envs

    def get_group_name(self) -> str:
        return ':cooperative'

    def get_repositories_list(self, ctx: ExecutionContext) -> Dict[str, str]:
        try:
            repos = ctx.get_arg_or_env('--repositories').split(',')
            repos_with_branch = {}

            for repo in repos:
                parts = repo.split('@@')
                repos_with_branch[parts[0]] = parts[1] if len(parts) >= 2 else 'master'

            return repos_with_branch

        except MissingInputException:
            self.io().warn('No repositories specified')
            return {}


class CooperativeSyncTask(BaseCooperativeTask):
    """Synchronize repositories"""

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--repositories', help='List of urls to repositories, comma separated')

    def get_name(self) -> str:
        return ':sync'

    def execute(self, ctx: ExecutionContext) -> bool:
        end_result = True

        for repository, branch in self.get_repositories_list(ctx).items():
            self.io().info('Syncing repository "%s"' % repository)

            if not self.sync_repository(repository, branch):
                self.io().error('Failed to synchronize repository "%s"' % repository)

                end_result = False
                continue

            self.io().info('Repository synced.')

        return end_result

    def sync_repository(self, git_url: str, branch: str):
        repository_dir = REPOSITORIES_DIRECTORY + '/' + self.extract_repository_name_from_git_url(git_url)

        try:
            if not os.path.isdir(repository_dir + '/'):
                self.sh(' '.join(['mkdir', '-p', repository_dir]))
                self.sh(' '.join(['git', 'clone', git_url, repository_dir]))
                self.sh('cd "%s" && git config advice.detachedHead false' % repository_dir)
                self.sh('cd "%s" && git checkout "%s"' % (repository_dir, branch))
                return True

            # pull the existing repository
            self.sh('''cd "%s" && git reset --hard HEAD && git checkout %s && git pull origin %s''' % (
                repository_dir, branch, branch
            ))

        except subprocess.CalledProcessError as e:
            self.io().error('Error fetching a git repository: %s' % str(e))
            return False

        return True

    @staticmethod
    def extract_repository_name_from_git_url(git_url: str) -> str:
        if git_url.startswith('http'):
            return urlparse(git_url).path[1:]

        if not git_url:
            return ''

        if os.path.isdir(git_url):
            return os.path.basename(git_url)

        separated = git_url[5:].split(':')

        if len(separated) < 2:
            raise Exception('Malformed GIT url "%s" in repositories list' % git_url)

        name = separated[1]
        parts = name.split('/')

        # support for custom GIT-SSH port (example: ssh://git@github.com:5000/riotkit-org/riotkit-do.git)
        if parts[0].isdecimal():
            name = '/'.join(parts[1:])

        if name.endswith('.git'):
            return name[0:-4]

        return name


class CooperativeInstallTask(BaseCooperativeTask):
    """Installs a snippet from the previously synchronized repository"""

    def get_name(self) -> str:
        return ':install'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('name', help='Snippet name')

    def execute(self, ctx: ExecutionContext) -> bool:
        name = ctx.get_arg('name')

        category_name, pkg_name = self.extract_category_and_pkg_names(name)
        path = self.find_snippet_path(pkg_name, category_name)

        if not path:
            self.io().error('Snippet not found in any synchronized repository. ' +
                            'Did you forget to do :cooperative:sync?')
            return False

        self.io().info('Installing snippet from "%s"' % path)

        # mock rkd_path, so the snippet can override the tasks
        rkd_path = os.getenv('RKD_PATH', '')
        snippet_rkd_path = os.path.realpath('./' + path + '/.rkd')

        if snippet_rkd_path:
            os.putenv('RKD_PATH', (rkd_path + ':' + snippet_rkd_path).strip(':'))

        try:
            subprocess.check_call(['rkd', ':snippet:wizard', path])
            subprocess.check_call(['rkd', ':snippet:install', path])
        finally:
            if os.path.isfile('.rkd/tmp-wizard.json'):
                os.unlink('.rkd/tmp-wizard.json')

            os.putenv('RKD_PATH', rkd_path)

        return True

    def extract_category_and_pkg_names(self, name: str) -> Tuple[str, str]:
        parts = name.split('/')
        category_name = parts[0] if len(parts) >= 2 else ''
        pkg_name = '/'.join(parts[1:]) if len(parts) >= 2 else name

        self.io().debug('category_name=%s, pkg_name=%s' % (category_name, pkg_name))

        return category_name, pkg_name

    def find_snippet_path(self, name: str, category_name: str):
        """Finds a snippet path by name.

        Raises:
            Exception: When snippet exists in multiple repositories
        """

        found_path = None

        for snippet_path in self.list_snippets(category_name):
            snippet_name = os.path.basename(snippet_path)

            if snippet_name == name:
                if found_path is not None:
                    self.io().error(('Ambiguous match, %s exists in %s and in %s. ' +
                                    'Maybe try to prepend a category name if it is available?') %
                                    (name, found_path, snippet_path))
                    return ''

                found_path = snippet_path

        return found_path

    def list_snippets(self, category_name: str):
        inside_path = category_name + '/' if category_name else ''
        glob_path = '.rkd/cooperative/**/snippets/' + inside_path + '**/snippet.json'

        self.io().debug('glob_path=%s' % glob_path)
        dirs = glob(glob_path, recursive=True)

        return list(set(map(lambda name: os.path.dirname(name), dirs)))


class CooperativeSnippetWizardTask(BaseCooperativeTask):
    """Snippet wizard - to be overridden by custom task provided by snippet"""

    def get_group_name(self) -> str:
        return ':snippet'

    def get_name(self) -> str:
        return ':wizard'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('path', help='Path to the root directory of the snippet files')

    def execute(self, ctx: ExecutionContext) -> bool:
        return True


class CooperativeSnippetInstallTask(BaseCooperativeTask):
    """Snippet installation - to be overridden by custom task provided by snippet"""

    def get_group_name(self) -> str:
        return ':snippet'

    def get_name(self) -> str:
        return ':install'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('path', help='Path to the root directory of the snippet files')

    def execute(self, ctx: ExecutionContext) -> bool:
        wizard = Wizard(self)
        wizard.load_previously_stored_values()
        os.environ.update(wizard.answers)

        self.rkd([
            ':j2:directory-to-directory',
            '--source="%s"' % ctx.get_arg('path') + '/files/',
            '--target="./"',
            '--pattern="(.*)"',
            '--copy-not-matching-files',
            '--template-filenames'
        ])
        return True



import os
import subprocess
from tempfile import TemporaryDirectory
from rkd.inputoutput import BufferedSystemIO
from rkd_cooperative import CooperativeInstallTask
from rkd_cooperative.test import BaseTestCase

TESTS_DIR = os.path.dirname(os.path.realpath(__file__))


class CooperativeInstallTaskTest(BaseTestCase):
    def test_functional_is_installing_dummy_snippet(self):
        """Functional test for installing a snippet from offline repository - successful case"""

        with TemporaryDirectory() as tmp_dir:
            # prepare
            os.chdir(tmp_dir)
            subprocess.check_call(['cp', '-pr', TESTS_DIR + '/examples/.rkd', '.rkd'])
            subprocess.check_call(['mkdir', '-p', '.rkd/cooperative'])
            subprocess.check_call(['cp', '-pr', TESTS_DIR + '/examples/repository-1', '.rkd/cooperative/repository-1'])

            os.environ['__WIZARD_INPUT'] = 'iwa-ait.org'

            task: CooperativeInstallTask = CooperativeInstallTask()
            self.satisfy_task_dependencies(task=task, io=BufferedSystemIO())

            task.execute(self.mock_execution_context(task, args={'name': 'nginx'}))

            with open('nginx.conf', 'r') as test_rendered_config_file:
                content = test_rendered_config_file.read()

                self.assertIn('server_name  iwa-ait.org;', content)

    def test_functional_trying_to_install_non_existing_snippet(self):
        with TemporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)

            io = BufferedSystemIO()
            task: CooperativeInstallTask = CooperativeInstallTask()
            self.satisfy_task_dependencies(task=task, io=io)

            task.execute(self.mock_execution_context(task, args={'name': 'some-non-existing-snippet'}))
            self.assertIn('Snippet not found in any synchronized repository. Did you forget to do :cooperative:sync?',
                          io.get_value())

    def test_functional_ambiguous_match_will_be_reported_when_there_are_two_packages_of_same_name(self):
        """Test that when same package name is in multiple repositories, then no package would be installed
        and the error would be shown instead"""

        with TemporaryDirectory() as tmp_dir:
            os.chdir(tmp_dir)
            subprocess.check_call(['cp', '-pr', TESTS_DIR + '/examples/.rkd', '.rkd'])
            subprocess.check_call(['mkdir', '-p', '.rkd/cooperative'])
            subprocess.check_call(['cp', '-pr', TESTS_DIR + '/examples/repository-1', '.rkd/cooperative/repository-1'])

            # this copy will duplicate all snippets, that's the key of our test
            subprocess.check_call(['cp', '-pr', TESTS_DIR + '/examples/repository-1', '.rkd/cooperative/repository-copy'])

            io = BufferedSystemIO()
            task: CooperativeInstallTask = CooperativeInstallTask()
            self.satisfy_task_dependencies(task=task, io=io)

            task.execute(self.mock_execution_context(task, args={'name': 'nginx'}))
            self.assertIn('Ambiguous match, nginx exists in', io.get_value())
            self.assertIn('.rkd/cooperative/repository-1/snippets/nginx', io.get_value())
            self.assertIn('.rkd/cooperative/repository-copy/snippets/nginx', io.get_value())

    def test_extract_category_and_pkg_names_returns_package_name_and_category_name_when_those_are_given(self):
        task = CooperativeInstallTask()
        task._io = BufferedSystemIO()
        category_name, pkg_name = task.extract_category_and_pkg_names('harbor/redis')

        self.assertEqual('harbor', category_name)
        self.assertEqual('redis', pkg_name)

    def test_extract_category_and_pkg_names_returns_only_package_name_when_category_is_not_present(self):
        task = CooperativeInstallTask()
        task._io = BufferedSystemIO()
        category_name, pkg_name = task.extract_category_and_pkg_names('redis')

        self.assertEqual('', category_name)
        self.assertEqual('redis', pkg_name)

    def test_extract_category_and_pkg_names_returns_multiple_directories_in_pkg_name(self):
        task = CooperativeInstallTask()
        task._io = BufferedSystemIO()
        category_name, pkg_name = task.extract_category_and_pkg_names('harbor/db/kv/redis')

        self.assertEqual('harbor', category_name)
        self.assertEqual('db/kv/redis', pkg_name)



"""
    RKD Snippet Cooperative
    =======================

    Autonomous RKD distribution dedicated for creating configuration sharing mechanisms
    (often called cooperatives, stores etc.)

"""

import os
from rkd.api.syntax import TaskDeclaration
from rkd import main as rkd_main
from rkd.standardlib.env import GetEnvTask
from rkd.standardlib.env import SetEnvTask
from rkd.standardlib.jinja import RenderDirectoryTask
from rkd.standardlib.jinja import FileRendererTask
from .tasks import CooperativeSnippetWizardTask
from .tasks import CooperativeSnippetInstallTask
from .tasks import CooperativeInstallTask
from .tasks import CooperativeSyncTask


def imports():
    return [
        TaskDeclaration(CooperativeSyncTask()),
        TaskDeclaration(CooperativeSnippetInstallTask()),
        TaskDeclaration(CooperativeInstallTask()),
        TaskDeclaration(CooperativeSnippetWizardTask()),
        TaskDeclaration(CooperativeSyncTask()),

        TaskDeclaration(GetEnvTask()),
        TaskDeclaration(SetEnvTask()),
        TaskDeclaration(FileRendererTask()),
        TaskDeclaration(RenderDirectoryTask())
    ]


def env_or_default(env_name: str, default: str):
    return os.environ[env_name] if env_name in os.environ else default


def main():
    os.environ['RKD_PATH'] = os.path.dirname(os.path.realpath(__file__)) + '/internal:' + os.getenv('RKD_PATH', '')
    os.environ['RKD_WHITELIST_GROUPS'] = env_or_default('RKD_WHITELIST_GROUPS', ':env,:cooperative')
    os.environ['RKD_ALIAS_GROUPS'] = env_or_default('RKD_ALIAS_GROUPS', '')
    os.environ['RKD_UI'] = env_or_default('RKD_UI', 'false')
    rkd_main()


if __name__ == '__main__':
    main()
